import docker
from typing import Optional, Union
from kubernetes import client as k8s_client
from llm_sandbox.const import SupportedLanguage
from llm_sandbox.docker import SandboxDockerSession
from llm_sandbox.kubernetes import SandboxKubernetesSession


class SandboxSession:
    def __new__(
        cls,
        client: Union[docker.DockerClient, k8s_client.CoreV1Api] = None,
        image: Optional[str] = None,
        dockerfile: Optional[str] = None,
        lang: str = SupportedLanguage.PYTHON,
        keep_template: bool = False,
        verbose: bool = False,
        use_kubernetes: bool = False,
        kube_namespace: Optional[str] = "default",
    ):
        """
        Create a new sandbox session
        :param client: Either Docker or Kubernetes client, if not provided, a new client will be created based on local context
        :param image: Docker image to use
        :param dockerfile: Path to the Dockerfile, if image is not provided
        :param lang: Language of the code
        :param keep_template: if True, the image and container will not be removed after the session ends
        :param verbose: if True, print messages (default is True)
        :param use_kubernetes: if True, use Kubernetes instead of Docker (default is False)
        :param kube_namespace: Kubernetes namespace to use (only if 'use_kubernetes' is True), default is 'default'
        """
        if use_kubernetes:
            return SandboxKubernetesSession(
                client=client,
                image=image,
                dockerfile=dockerfile,
                lang=lang,
                keep_template=keep_template,
                verbose=verbose,
                kube_namespace=kube_namespace,
            )

        return SandboxDockerSession(
            client=client,
            image=image,
            dockerfile=dockerfile,
            lang=lang,
            keep_template=keep_template,
            verbose=verbose,
        )



import docker
import docker.errors
from typing import Optional

from docker import DockerClient
from llm_sandbox.const import SupportedLanguage


def image_exists(client: DockerClient, image: str) -> bool:
    """
    Check if a Docker image exists
    :param client: Docker client
    :param image: Docker image
    :return: True if the image exists, False otherwise
    """
    try:
        client.images.get(image)
        return True
    except docker.errors.ImageNotFound:
        return False
    except Exception as e:
        raise e


def get_libraries_installation_command(lang: str, library: str) -> Optional[str]:
    """
    Get the command to install libraries for the given language
    :param lang: Programming language
    :param library: List of libraries
    :return: Installation command
    """
    if lang == SupportedLanguage.PYTHON:
        return f"pip install {library}"
    elif lang == SupportedLanguage.JAVA:
        return f"mvn install:install-file -Dfile={library}"
    elif lang == SupportedLanguage.JAVASCRIPT:
        return f"yarn add {library}"
    elif lang == SupportedLanguage.CPP:
        return f"apt-get install {library}"
    elif lang == SupportedLanguage.GO:
        return f"go get -u {library}"
    elif lang == SupportedLanguage.RUBY:
        return f"gem install {library}"
    else:
        raise ValueError(f"Language {lang} is not supported")


def get_code_file_extension(lang: str) -> str:
    """
    Get the file extension for the given language
    :param lang: Programming language
    :return: File extension
    """
    if lang == SupportedLanguage.PYTHON:
        return "py"
    elif lang == SupportedLanguage.JAVA:
        return "java"
    elif lang == SupportedLanguage.JAVASCRIPT:
        return "js"
    elif lang == SupportedLanguage.CPP:
        return "cpp"
    elif lang == SupportedLanguage.GO:
        return "go"
    elif lang == SupportedLanguage.RUBY:
        return "rb"
    else:
        raise ValueError(f"Language {lang} is not supported")


def get_code_execution_command(lang: str, code_file: str) -> list:
    """
    Return the execution command for the given language and code file.
    :param lang: Language of the code
    :param code_file: Path to the code file
    :return: List of execution commands
    """
    if lang == SupportedLanguage.PYTHON:
        return [f"python {code_file}"]
    elif lang == SupportedLanguage.JAVA:
        return [f"java {code_file}"]
    elif lang == SupportedLanguage.JAVASCRIPT:
        return [f"node {code_file}"]
    elif lang == SupportedLanguage.CPP:
        return [f"g++ -o a.out {code_file}", "./a.out"]
    elif lang == SupportedLanguage.GO:
        return [f"go run {code_file}"]
    elif lang == SupportedLanguage.RUBY:
        return [f"ruby {code_file}"]
    else:
        raise ValueError(f"Language {lang} is not supported")



from abc import ABC, abstractmethod
from typing import Optional, List


class ConsoleOutput:
    def __init__(self, text: str):
        self._text = text

    @property
    def text(self):
        return self._text

    def __str__(self):
        return f"ConsoleOutput(text={self.text})"


class KubernetesConsoleOutput(ConsoleOutput):
    def __init__(self, exit_code: int, text: str):
        super().__init__(text)
        self.exit_code = exit_code

    def __str__(self):
        return f"KubernetesConsoleOutput(text={self.text}, exit_code={self.exit_code})"


class Session(ABC):
    def __init__(self, lang: str, verbose: bool = True, *args, **kwargs):
        self.lang = lang
        self.verbose = verbose
        super().__init__(*args, **kwargs)

    @abstractmethod
    def open(self):
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError

    @abstractmethod
    def run(self, code: str, libraries: Optional[List] = None) -> ConsoleOutput:
        raise NotImplementedError

    @abstractmethod
    def copy_to_runtime(self, src: str, dest: str):
        raise NotImplementedError

    @abstractmethod
    def copy_from_runtime(self, src: str, dest: str):
        raise NotImplementedError

    @abstractmethod
    def execute_command(self, command: str):
        raise NotImplementedError

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args, **kwargs):
        self.close()



import io
import os
import time
import uuid
import tarfile
from typing import List, Optional

from kubernetes import client as k8s_client, config
from kubernetes.stream import stream
from llm_sandbox.base import Session, ConsoleOutput, KubernetesConsoleOutput
from llm_sandbox.utils import (
    get_libraries_installation_command,
    get_code_file_extension,
    get_code_execution_command,
)
from llm_sandbox.const import SupportedLanguage, SupportedLanguageValues, DefaultImage


class SandboxKubernetesSession(Session):
    def __init__(
        self,
        client: Optional[k8s_client.CoreV1Api] = None,
        image: Optional[str] = None,
        dockerfile: Optional[str] = None,
        lang: str = SupportedLanguage.PYTHON,
        keep_template: bool = False,
        verbose: bool = False,
        kube_namespace: Optional[str] = "default",
    ):
        """
        Create a new sandbox session
        :param client: Kubernetes client, if not provided, a new client will be created based on local Kubernetes context
        :param image: Docker image to use
        :param dockerfile: Path to the Dockerfile, if image is not provided
        :param lang: Language of the code
        :param keep_template: if True, the image and container will not be removed after the session ends
        :param verbose: if True, print messages
        :param kube_namespace: Kubernetes namespace to use, default is 'default'
        """
        super().__init__(lang, verbose)
        if lang not in SupportedLanguageValues:
            raise ValueError(
                f"Language {lang} is not supported. Must be one of {SupportedLanguageValues}"
            )

        if not image:
            image = DefaultImage.__dict__[lang.upper()]

        if not client:
            if self.verbose:
                print("Using local Kubernetes context since client is not provided..")

            config.load_kube_config()
            self.client = k8s_client.CoreV1Api()
        else:
            self.client = client

        self.image = image
        self.kube_namespace = kube_namespace
        self.pod_name = f"sandbox-{lang.lower()}-{uuid.uuid4().hex}"
        self.keep_template = keep_template
        self.container = None

    def open(self):
        self._create_kubernetes_pod()

    def _create_kubernetes_pod(self):
        pod_manifest = {
            "apiVersion": "v1",
            "kind": "Pod",
            "metadata": {
                "name": self.pod_name,
                "namespace": self.kube_namespace,
                "labels": {"app": "sandbox"},
            },
            "spec": {
                "containers": [
                    {"name": "sandbox-container", "image": self.image, "tty": True}
                ]
            },
        }
        self.client.create_namespaced_pod(
            namespace=self.kube_namespace, body=pod_manifest
        )

        while True:
            pod = self.client.read_namespaced_pod(
                name=self.pod_name, namespace=self.kube_namespace
            )
            if pod.status.phase == "Running":
                break
            time.sleep(1)

        self.container = self.pod_name

    def close(self):
        self._delete_kubernetes_pod()

    def _delete_kubernetes_pod(self):
        self.client.delete_namespaced_pod(
            name=self.pod_name,
            namespace=self.kube_namespace,
            body=k8s_client.V1DeleteOptions(),
        )

    def run(self, code: str, libraries: Optional[List] = None) -> ConsoleOutput:
        if not self.container:
            raise RuntimeError(
                "Session is not open. Please call open() method before running code."
            )

        if libraries:
            if self.lang == SupportedLanguage.GO:
                self.execute_command("mkdir -p /example")
                self.execute_command("go mod init example", workdir="/example")
                self.execute_command("go mod tidy", workdir="/example")

                for library in libraries:
                    install_command = get_libraries_installation_command(
                        self.lang, library
                    )
                    output = self.execute_command(install_command, workdir="/example")
                    if output.exit_code != 0:
                        raise RuntimeError(
                            f"Failed to install library {library}: {output}"
                        )
            else:
                for library in libraries:
                    install_command = get_libraries_installation_command(
                        self.lang, library
                    )
                    output = self.execute_command(install_command)
                    if output.exit_code != 0:
                        raise RuntimeError(
                            f"Failed to install library {library}: {output}"
                        )

        code_file = f"/tmp/code.{get_code_file_extension(self.lang)}"
        if self.lang == SupportedLanguage.GO:
            code_dest_file = "/example/code.go"
        else:
            code_dest_file = code_file

        with open(code_file, "w") as f:
            f.write(code)

        self.copy_to_runtime(code_file, code_dest_file)
        commands = get_code_execution_command(self.lang, code_dest_file)

        output = KubernetesConsoleOutput(0, "")
        for command in commands:
            if self.lang == SupportedLanguage.GO:
                output = self.execute_command(command, workdir="/example")
            else:
                output = self.execute_command(command)

            if output.exit_code != 0:
                break

        return ConsoleOutput(output.text)

    def copy_to_runtime(self, src: str, dest: str):
        if not self.container:
            raise RuntimeError(
                "Session is not open. Please call open() method before copying files."
            )

        start_time = time.time()
        if self.verbose:
            print(f"Copying {src} to {self.container}:{dest}..")

        dest_dir = os.path.dirname(dest)
        dest_file = os.path.basename(dest)

        if dest_dir:
            self.execute_command(f"mkdir -p {dest_dir}")

        with open(src, "rb") as f:
            tarstream = io.BytesIO()
            with tarfile.open(fileobj=tarstream, mode="w") as tar:
                tarinfo = tarfile.TarInfo(name=dest_file)
                tarinfo.size = os.path.getsize(src)
                tar.addfile(tarinfo, f)
            tarstream.seek(0)

            exec_command = ["tar", "xvf", "-", "-C", dest_dir]
            resp = stream(
                self.client.connect_get_namespaced_pod_exec,
                self.container,
                self.kube_namespace,
                command=exec_command,
                stderr=True,
                stdin=True,
                stdout=True,
                tty=False,
                _preload_content=False,
            )
            while resp.is_open():
                resp.update(timeout=1)
                if resp.peek_stdout():
                    print(resp.read_stdout())
                if resp.peek_stderr():
                    print(resp.read_stderr())
                resp.write_stdin(tarstream.read(4096))
            resp.close()

        end_time = time.time()
        if self.verbose:
            print(
                f"Copied {src} to {self.container}:{dest} in {end_time - start_time:.2f} seconds"
            )

    def copy_from_runtime(self, src: str, dest: str):
        if not self.container:
            raise RuntimeError(
                "Session is not open. Please call open() method before copying files."
            )

        if self.verbose:
            print(f"Copying {self.container}:{src} to {dest}..")

        exec_command = ["tar", "cf", "-", src]
        resp = stream(
            self.client.connect_get_namespaced_pod_exec,
            self.container,
            self.kube_namespace,
            command=exec_command,
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
            _preload_content=False,
        )
        with open(dest, "wb") as f:
            while resp.is_open():
                resp.update(timeout=1)
                if resp.peek_stdout():
                    f.write(resp.read_stdout())
                if resp.peek_stderr():
                    print(resp.read_stderr())

    def execute_command(
        self, command: str, workdir: Optional[str] = None
    ) -> KubernetesConsoleOutput:
        if not self.container:
            raise RuntimeError(
                "Session is not open. Please call open() method before executing commands."
            )

        if self.verbose:
            print(f"Executing command: {command}")

        if workdir:
            exec_command = ["sh", "-c", f"cd {workdir} && {command}"]
        else:
            exec_command = ["/bin/sh", "-c", command]

        resp = stream(
            self.client.connect_get_namespaced_pod_exec,
            self.container,
            self.kube_namespace,
            command=exec_command,
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
            _preload_content=False,
        )

        output = ""
        if self.verbose:
            print("Output:", end=" ")

        while resp.is_open():
            resp.update(timeout=1)
            if resp.peek_stdout():
                chunk = resp.read_stdout()
                output += chunk
                if self.verbose:
                    print(chunk, end="")
            if resp.peek_stderr():
                chunk = resp.read_stderr()
                output += chunk
                if self.verbose:
                    print(chunk, end="")

        exit_code = resp.returncode
        return KubernetesConsoleOutput(exit_code, output)
